/*
 * Crash Bandicoot (1996) Loading Screen Converter v2.1
 * Copyright (C) 2017  b122251
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* File Inclusions */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "cbload.h"
#include "cblcints.h"
#include "imgtools.h"

/*
 * Main function of the program.
 * Parameters:
 *   argc = Number of command-line parameters
 *   argv = Command-line parameters
 * Return Values:
 *   0 = Everything went well
 *   1 = NSD-file could not be opened
 *   2 = BMP-file could not be opened
 *   3 = Reading from NSD-file failed
 *   4 = Writing to BMP-file failed
 *   5 = Buffer could not be allocated
 *   6 = NSD-file does not contain an image
 *   7 = BMP-file is not 432x144 pixels in size
 *   8 = BMP-file is not 4-bit or 8-bit indexed colour
 *   9 = BMP-file is not valid
 *   10 = Reading from BMP-file failed
 *   11 = Writing to NSD-file failed
 */
int main(int argc, char *argv[]){
	/* Variable Declarations */
	int retval; /* Return value for the program */
	
	/* Interprets command-line parameters */
	retval=interpretParameters(argc,argv);
	
	/* Prints error messages if needed */
	if (retval!=0) printErrors(argv,retval);
	
	/* Returns potential error numbers and closes the program */
	return retval;
}

/*
 * Function that interprets user input and calls appropriate functions.
 * Parameters:
 *   argc = Number of command-line parameters
 *   argv = Command-line parameters
 * Return Values:
 *   -1 = Insufficient parameters
 *    0 = Everything went well
 *   >0 = Error code of called functions
 */
static int interpretParameters(int argc, char *argv[]){
	/* Variable Declarations */
	int retval=-1; /* Return value of this function */
	size_t curchar; /* Current character when converting parameters */
	
	/* Checks number of parameters and length of argv[2] */
	if (argc<4) return -1;
	if (strlen(argv[2])<3) return -1;
	
	/* Converts argv[1] to lower case */
	for (curchar=0;curchar<strlen(argv[1]);curchar++)
		argv[1][curchar]=tolower(argv[1][curchar]);
	
	/* Interprets command parameter */
	if (strncmp(argv[1],"extract",7)==0){
		retval=nsdToBmp(argv[2],argv[3]);
	} else if (strncmp(argv[1],"replace",7)==0){
		retval=bmpToNsd(argv[2],argv[3]);
	}
	
	return retval;
}

/*
 * Function that extracts images from Crash Bandicoot (1996) NSD-files
 * Parameters:
 *   nsdPath = Path to input NSD-file
 *   bmpPath = Path to output BMP-file
 * Return Values:
 *   0 = Everything went well
 *   1 = NSD-file could not be opened
 *   2 = BMP-file could not be opened
 *   3 = Reading from NSD-file failed
 *   4 = Writing to BMP-file failed
 *   5 = Buffer could not be allocated
 *   6 = NSD-file does not contain an image
 */
static int nsdToBmp(char *nsdPath, char *bmpPath){
	/* Variable Declarations */
	FILE *nsd=NULL; /* Input NSD-file */
	FILE *bmp=NULL; /* Output BMP-file */
	char *buffer=NULL; /* Buffer */
	long int nsdsize; /* Length of input NSD-file in bytes */
	long int nsdpos; /* Current position in NSD-file */
	size_t buffersize; /* Length of buffer in bytes */
	size_t freesizet1; /* size_t for general use */
	int freeint1; /* Integers for general use */
	unsigned INT16 freeushort1; /* Integer for general use */
	long int freelong1; /* Long integer for general use */
	int retval=0; /* Return value of this function */
	
	/* Opens input file */
	nsd=fopen(nsdPath,"rb");
	if (nsd==NULL){
		retval=1;
		goto end;
	}
	
	/* Determines nsdsize */
	freeint1=fseek(nsd,0,SEEK_END);
	if (freeint1!=0){
		retval=3;
		goto end;
	}
	nsdsize=ftell(nsd);
	
	/* Determines if input NSD-file contains an image */
	if (nsdsize<62720){
		retval=6;
		goto end;
	}
	
	/* Opens output file */
	bmp=fopen(bmpPath,"wb");
	if (bmp==NULL){
		retval=2;
		goto end;
	}
	
	/* Writes bitmap header */
	freeint1=bmpheader(bmp,63286,1078,432,144,8,0,62208,0,0,256,256);
	if (freeint1!=0){
		retval=(freeint1*2);
		goto end;
	}
	
	/* Allocates buffer */
	buffer=NULL;
	buffersize=0x7FFF;
	do{
		buffer=malloc((size_t) buffersize);
		if ((buffersize<=432)&&(buffer==NULL)){
			retval=5;
			goto end;
		}
		if (buffer==NULL) buffersize>>=1;
	}while (buffer==NULL);
	buffersize-=(buffersize%432);
	
	/* Processes and writes palette */
	freeint1=fseek(nsd,(nsdsize-62720),SEEK_SET);
	if (freeint1!=0){
		retval=3;
		goto end;
	}
	memset(buffer,0,4);
	for (freeint1=0;freeint1<=255;freeint1++){
		freesizet1=fread(&freeushort1,1,2,nsd);
		if (freesizet1!=2){
			retval=3;
			goto end;
		}
		buffer[2]=(char)((((unsigned INT16)(freeushort1&0x1F))<<3)+
		                 (((unsigned INT16)(freeushort1&0x1C))>>2));
		buffer[1]=(char)((((unsigned INT16)(freeushort1&0x3E0))>>2)+
		                 (((unsigned INT16)(freeushort1&0x380))>>7));
		buffer[0]=(char)((((unsigned INT16)(freeushort1&0x7C00))>>7)+
		                 (((unsigned INT16)(freeushort1&0x7000))>>12));
		freesizet1=fwrite(buffer,1,4,bmp);
		if (freesizet1!=4){
			retval=4;
			goto end;
		}
	}
	
	/* Processes image data (backwards) in chunks */
	freelong1=(nsdsize-62208);
	nsdpos=nsdsize;
	while (nsdpos>freelong1){
		/* Adjusts the size of the final chunk if needed */
		if ((nsdpos-freelong1)<(long int)buffersize) 
			buffersize=(size_t)(nsdpos-freelong1);
		
		/* Reads chunk into buffer */
		freeint1=fseek(nsd,(INT32) (nsdpos-buffersize),SEEK_SET);
		if (freeint1!=0){
			retval=3;
			goto end;
		}
		freesizet1=fread(buffer,1,(size_t) buffersize,nsd);
		if (freesizet1!=(size_t) buffersize){
			retval=3;
			goto end;
		}
		
		/* Flips chunk vertically */
		freeint1=verticalFlip(buffer,432,(size_t)(buffersize/432));
		if (freeint1!=0){
			retval=5;
			goto end;
		}
		
		/* Writes chunk to output BMP-file */
		freesizet1=fwrite(buffer,1,(size_t) buffersize,bmp);
		if (freesizet1!=(size_t) buffersize){
			retval=4;
			goto end;
		}
		
		nsdpos-=buffersize;
	}
	
	/* Closes files, frees buffer and returns function */
end:
	if (buffer!=NULL) free(buffer);
	if (nsd!=NULL){
		freeint1=fclose(nsd);
		if (freeint1!=0) retval=2;
	}
	if (bmp!=NULL){
		freeint1=fclose(bmp);
		if (freeint1!=0) retval=4;
	}
	return retval;
}

/*
 * Function that replaces images from Crash Bandicoot (1996) NSD-files
 * Parameters:
 *   nsdPath = Path to input BMP-file
 *   bmpPath = Path to output NSD-file
 * Return Values:
 *   0 = Everything went well
 *   1 = NSD-file could not be opened
 *   2 = BMP-file could not be opened
 *   3 = Reading from NSD-file failed
 *   4 = Writing to BMP-file failed
 *   5 = Buffer could not be allocated
 *   6 = NSD-file does not contain an image
 *   7 = BMP-file is not 432x144 pixels in size
 *   8 = BMP-file is not 4-bit or 8-bit indexed colour
 *   9 = BMP-file is not valid
 *   10 = Reading from BMP-file failed
 *   11 = Writing to NSD-file failed
 */
static int bmpToNsd(char *nsdPath, char *bmpPath){
	/* Variable Declarations */
	FILE *bmp=NULL; /* Input BMP-file */
	FILE *nsd=NULL; /* Output NSD-file */
	unsigned char *buffer=NULL; /* Buffer */
	long int bmpsize; /* Length of input BMP-file in bytes */
	long int bmppos; /* Current position in BMP-file */
	long int nsdsize; /* Length of output NSD-file in bytes */
	struct bmpinfo ininfo; /* Input BMP header data */
	size_t buffersize; /* Length of buffer in bytes */
	size_t freesizet1; /* size_t for general use */
	long int freelong1, freelong2; /* Long integers for general use */
	int freeint1; /* Integers for general use */
	unsigned INT16 freeushort1; /* Integer for general use */
	int retval=0; /* Return value of this function */
	
	/* Opens input file */
	bmp=fopen(bmpPath,"rb");
	if (bmp==NULL){
		retval=2;
		goto end;
	}
	
	/* Determines bmpsize */
	freeint1=fseek(bmp,0,SEEK_END);
	if (freeint1!=0){
		retval=10;
		goto end;
	}
	bmpsize=ftell(bmp);
	
	/* Opens output file */
	nsd=fopen(nsdPath,"r+b");
	if (nsd==NULL){
		retval=1;
		goto end;
	}
	
	/* Determines nsdsize */
	freeint1=fseek(nsd,0,SEEK_END);
	if (freeint1!=0){
		retval=11;
		goto end;
	}
	nsdsize=ftell(nsd);
	
	/* Determines if output NSD-file contains an image */
	if (nsdsize<62720){
		retval=6;
		goto end;
	}
	
	/* Reads bitmap header */
	ininfo=readBMPHeader(bmp);
	if (ininfo.complete==0){
		retval=10;
		goto end;
	}
	
	/* Checks BMP dimensions */
	if (ininfo.width!=432){
		retval=7;
		goto end;
	}
	if (ininfo.height!=144){
		retval=7;
		goto end;
	}
	
	/* Checks bpp */
	if ((ininfo.bpp!=4)&&(ininfo.bpp!=8)){
		retval=8;
		goto end;
	}
	
	/* Checks number of colours */
	if ((ininfo.colours<3)||(ininfo.colours>256)){
		retval=8;
		goto end;
	}
	
	/* Checks whether BMP-header is valid */
	if (bmpsize<((long int)(ininfo.imageOffset+
	                        ((ininfo.bpp==8)?62208:31104)))){
		retval=8;
		goto end;
	}
	
	/* Allocates buffer */
	buffer=NULL;
	buffersize=0x7FFF;
	do{
		buffer=malloc((size_t) buffersize);
		if ((buffersize<=432)&&(buffer==NULL)){
			retval=5;
			goto end;
		}
		if (buffer==NULL) buffersize>>=1;
	}while (buffer==NULL);
	buffersize-=(buffersize%432);
	
	/* Processes and writes palette */
	freelong1=(long int) ininfo.imageOffset-(ininfo.colours*4);
	freeint1=fseek(bmp,freelong1,SEEK_SET);
	if (freeint1!=0){
		retval=10;
		goto end;
	}
	freeint1=fseek(nsd,(nsdsize-62720),SEEK_SET);
	if (freeint1!=0){
		retval=11;
		goto end;
	}
	for (freeint1=(int) ininfo.colours;freeint1>0;freeint1--){
		freesizet1=fread(buffer,1,4,bmp);
		if (freesizet1!=4){
			retval=10;
			goto end;
		}
		if (buffer[0]<=(unsigned char)251) buffer[0]+=(unsigned char)4;
		else buffer[0]=(unsigned char)255;
		buffer[0]-=((buffer[0]-(buffer[0]%33))/(unsigned char)33);
		buffer[0]=((buffer[0]-(buffer[0]%8))/(unsigned char)8);
		if (buffer[1]<=(unsigned char)251) buffer[1]+=(unsigned char)4;
		else buffer[1]=(unsigned char)255;
		buffer[1]-=((buffer[1]-(buffer[1]%33))/(unsigned char)33);
		buffer[1]=((buffer[1]-(buffer[1]%8))/(unsigned char)8);
		if (buffer[2]<=(unsigned char)251) buffer[2]+=(unsigned char)4;
		else buffer[2]=(unsigned char)255;
		buffer[2]-=((buffer[2]-(buffer[2]%33))/(unsigned char)33);
		buffer[2]=((buffer[2]-(buffer[2]%8))/(unsigned char)8);
		freeushort1=(unsigned INT16) (0x8000|(buffer[2]&0x1F)|
		             ((buffer[1]&0x1F)<<5)|((buffer[0]&0x1F)<<10));
		freesizet1=fwrite(&freeushort1,1,2,nsd);
		if (freesizet1!=2){
			retval=11;
			goto end;
		}
	}
	
	/* Processes image data (backwards) in chunks */
	freeint1=fseek(nsd,(nsdsize-62208),SEEK_SET);
	if (freeint1!=0){
		retval=11;
		goto end;
	}
	bmppos=(long int)(ininfo.imageOffset+((ininfo.bpp==8)?62208:31104));
	while (bmppos>(long int)ininfo.imageOffset){
		/* Adjusts the size of the final chunk if needed */
		freelong1=((bmppos-ininfo.imageOffset)*((ininfo.bpp==8)?1:2));
		if (freelong1<(long int)buffersize) buffersize=(size_t)freelong1;
		
		/* Reads chunk into buffer */
		freelong1=(long int)(bmppos-(buffersize/((ininfo.bpp==8)?1:2)));
		freeint1=fseek(bmp,(INT32) freelong1,SEEK_SET);
		if (freeint1!=0){
			retval=10;
			goto end;
		}
		freelong1=(long int)(buffersize/((ininfo.bpp==8)?1:2));
		freesizet1=fread(buffer,1,(size_t) freelong1,bmp);
		if (freesizet1!=(size_t) freelong1){
			retval=10;
			goto end;
		}
		bmppos-=freelong1;
		
		/* Converts from 4-bit to 8-bit if necessary */
		if (ininfo.bpp==4){
			freelong1=(long int)(buffersize-1);
			freelong2=(long int)((buffersize/2)-1);
			for (;freelong2>=0;freelong2--){
				buffer[freelong1]=(buffer[freelong2]&0x0F);
				freelong1--;
				buffer[freelong1]=((buffer[freelong2]&0xF0)>>4);
				freelong1--;
			}
		}
		
		/* Flips chunk vertically */
		freeint1=verticalFlip((char *)buffer,432,
		                      (size_t)(buffersize/432));
		if (freeint1!=0){
			retval=5;
			goto end;
		}
		
		/* Writes chunk to output BMP-file */
		freesizet1=fwrite(buffer,1,(size_t) buffersize,nsd);
		if (freesizet1!=(size_t) buffersize){
			retval=11;
			goto end;
		}
	}
	
	/* Closes files, frees buffer and returns function */
end:
	if (buffer!=NULL) free(buffer);
	if (bmp!=NULL){
		freeint1=fclose(bmp);
		if (freeint1!=0) retval=10;
	}
	if (nsd!=NULL){
		freeint1=fclose(nsd);
		if (freeint1!=0) retval=11;
	}
	return retval;
}

/*
 * Writes error messages to the terminal
 * Paramters:
 *   argv = Command-line parameters
 *   errnum = Error number
 */
static void printErrors(char *argv[], int errnum){
	switch (errnum){
		case -1:
			printf("cbload: Missing file operands\nUsage: cbload "
			       "[extract/replace] [NSD-File] [BMP-File]\n");
			break;
		case 1:
			printf("cbload: %s could not be opened\n",argv[2]);
			break;
		case 2:
			printf("cbload: %s could not be opened\n",argv[3]);
			break;
		case 3:
			printf("cbload: Reading from %s failed\n",argv[2]);
			break;
		case 4:
			printf("cbload: Writing to %s failed\n",argv[3]);
			break;
		case 5:
			printf("cbload: Buffer could not be allocated\n");
			break;
		case 6:
			printf("cbload: %s does not contain an image\n",argv[2]);
			break;
		case 7:
			printf("cbload: %s is not 432x144 in size\n",argv[3]);
			break;
		case 8:
			printf("cbload: %s is not 4-bit or 8-bit indexed colour\n",argv[3]);
			break;
		case 9:
			printf("cbload: %s is not a valid bitmap-file\n",argv[3]);
			break;
		case 10:
			printf("cbload: Reading from %s failed\n",argv[3]);
			break;
		case 11:
			printf("cbload: Writing to %s failed\n",argv[2]);
			break;
	}
}
